from digcnv import digCNV_logger

import configparser


def getConfigFileExample(exmpl_conf_output: str):
    """Write an example config file used for next steps

    Args:
        exmpl_conf_output (str): pathway where the example config file will be written
    """    
    config_file = configparser.ConfigParser()
    config_file["Inputs"] = {
        "PC_output_path": "<Path to PennCNV file>",
        "PC_QC_path" : "<Path to PennCNV microarray quality file>",
        "QS_output_path": "<Path to QuantiSNP file>"
    }
    # config_file["CallRates"] = {
    #     "CallRate_path": "<Path to CallRates file>",
    #     "individual_colname": "<Column name of individials>",
    #     "callrates_colname": "<Column name of Callrate values>"
    # }
    # config_file["PFBs"] = {
    #     "PFB_path": "<Path to PFB file>"
    # }
    config_file["Output"] = {
        "Save_to_file": "<True or False if you want to save the dataset to a file>",
        "Output_path": "<Path of the future output file>"
    }

    config_file["DigCNV"] = {
        "model_path": "<Path of the downloaded model. Available at : https://murena.io/s/xEsyae6gxfMEnWJ>"
    }

    with open(exmpl_conf_output, 'w') as configfileObj:
        config_file.write(configfileObj)
        configfileObj.flush()
        configfileObj.close()
    digCNV_logger.logger.info("Example Config file written to {}".format(exmpl_conf_output))


def readDigCNVConfFile(conf_file_path: str) -> dict:
    """Reads the given config file and return a dictionnary of named parameters

    Args:
        conf_file_path (str): pathway of the config file

    Returns:
        dict: Dictionnary containing all parameters values
    """ 
    parameters = {}
    config_file = configparser.ConfigParser()
    config_file.read(conf_file_path)
    digCNV_logger.logger.info("Config file {} opened".format(conf_file_path))
    parameters["PC"] = config_file.get('Inputs', "PC_output_path")
    parameters["QS"] = config_file.get('Inputs', "QS_output_path")
    parameters["QC"] = config_file.get('Inputs', "PC_QC_path")

    parameters["CallRate"] = config_file.get('Inputs', "callrate_path")
    # parameters["ind_name"] = config_file.get('CallRates', "individual_colname")
    # parameters["CR_name"] = config_file.get('CallRates', "callrates_colname")

    # parameters["PFB"] = config_file.get('PFBs', "PFB_path")
    parameters['DigCnvModel'] = config_file.get('DigCNV', 'model_path')
    parameters["save"] = config_file.get('Output', 'Save_to_file')
    parameters["output"] = config_file.get('Output', 'Output_path')
    digCNV_logger.logger.info("Set of parameters created for DigCNV")
    return parameters



from digcnv.digCNV_logger import logger as dc_logger
import pandas as pd
import numpy as np
from os.path import exists, split, join


def addMicroArrayQualityData(cnvs: pd.DataFrame, data_path: str) -> pd.DataFrame:
    """Add Microarray quality data to the list of merged CNVs

    :param cnvs: list of CNVs and their quality scores
    :type cnvs: pd.DataFrame
    :param data_path: pathway containing PennCNV quality chip output
    :type data_path: str
    :return: list of CNVs with sample quality aggregated
    :rtype: pd.DataFrame
    """
    data = pd.read_csv(data_path, sep="\t")
    dc_logger.info("Micro-array quality data opened")
    dc_logger.info("Add {} columns".format(data.columns.tolist()))
    data.columns = ["SampleID", "LRR_mean", "LRR_median", "LRR_SD",
                    "BAF_mean", "BAF_median", "BAF_SD", "BAF_DRIFT", "WF", "GCWF"]
    cnvs_qc = pd.merge(cnvs, data, how="left", on='SampleID')
    dc_logger.info("Micro-array quality data merged to CNV quality data")
    return cnvs_qc


def addDerivedFeatures(cnvs: pd.DataFrame) -> pd.DataFrame:
    """Compute and add 3 derived CNV scores 

    :param cnvs: list of CNVs with at least 4 mandatory columns, (`START`, `STOP`, `SNP`, `SCORE`) 
    :type cnvs: pd.DataFrame
    :return: same list of CNVs with 3 new column scores, (`SIZE`, `DENSITY`, `Score_SNP`)
    :rtype: pd.DataFrame
    """
    cnvs["SIZE"] = cnvs["STOP"] - cnvs["START"] + 1
    cnvs["DENSITY"] = cnvs["SNP"] / cnvs["SIZE"]
    cnvs["Score_SNP"] = cnvs["SCORE"] / cnvs["SNP"]
    dc_logger.info("Derived features (DENSITY and Score_SNP columns) created")
    return cnvs


def addCallRateToDataset(cnvs: pd.DataFrame, call_rate_path: str, callrate_colname="callRate", individual_colname="FID") -> pd.DataFrame:
    """Add CallRate information to the given list of CNVs

    :param cnvs: list of CNVs with their scores
    :type cnvs: pd.DataFrame
    :param call_rate_path: Pathway to the file listing CallRate data of each microarray. The file mut be a **tsv**.
    :type call_rate_path: str
    :param callrate_colname: CallRate column name, defaults to "callRate"
    :type callrate_colname: str, optional
    :param individual_colname: Individual column name, defaults to "FID"
    :type individual_colname: str, optional
    :raises Exception: If the given CallRate pathway doesn't exist
    :raises Exception: If the given CallRate file contains more than one value by sampleID
    :return: list of CNVs with CallRates aggregated
    :rtype: pd.DataFrame
    """
    if exists(call_rate_path):
        callrates = pd.read_csv(call_rate_path, sep='\t')
    else:
        raise Exception("Given path doesn't exist")
    callrates.rename(columns={callrate_colname: "CallRate"}, inplace=True)
    if len(callrates[individual_colname].unique()) != callrates.shape[0]:
        raise Exception("CallRates file must contains only unique individuals")
    if "CallRate" in cnvs.columns.tolist():
        dc_logger.info("Clean existing CallRate column")
        cnvs.drop(columns={"CallRate"}, inplace=True)

    cnvs_with_callrate = pd.merge(cnvs, callrates.loc[:, [
                                  individual_colname, "CallRate"]], how="left", left_on=individual_colname, right_on=individual_colname)

    if individual_colname != "SampleID":
        cnvs_with_callrate.drop(columns=[individual_colname], inplace=True)
    dc_logger.info("CallRate added to dataset")
    return cnvs_with_callrate


def addNbProbeByTech(cnvs: pd.DataFrame, nb_prob_tech=None, pfb_file_path=None) -> pd.DataFrame:
    """Add the number of Probes used to genotyped individuals. **Warning**: if your data comes from multiple genotyping technologies
    Please split your dataframe to get only one technology by dataframe before using this function.

    :param cnvs: list of CNVs with their scores
    :type cnvs: pd.DataFrame
    :param nb_prob_tech: Number of probe in the genotyping technology used, defaults to None
    :type nb_prob_tech: int, optional
    :param pfb_file_path: Pathway to the PFB file to compute get the number of probe used by the technology, defaults to None
    :type pfb_file_path: str, optional
    raises Exception: If none of the optional value is given. You have to set either nb_probe_tech or pfb_file_path parameter.
    :return: list of CNVs with Nb probe used in technology column aggregated
    :rtype: pd.DataFrame
    """
    if nb_prob_tech != None:
        cnvs["Nb_Probe_tech"] = nb_prob_tech
        dc_logger.info(
            "Number of probes in technology added thanks to the given number")
    elif pfb_file_path != None:
        lines = 0
        with open(pfb_file_path) as f:
            for line in f:
                lines = lines + 1
        cnvs["Nb_Probe_tech"] = lines - 1
        dc_logger.info(
            "Number of probes in technology added after counting number of lines in pfb file")
    else:
        raise Exception(
            "You have to give at least one of these two parameters (nb_prob_tech or pfb_file_path)")

    dc_logger.info("Number of probes in technology added")
    return cnvs


def addChromosomicAnnotation(cnvs: pd.DataFrame, centromere_list_path=None, segdup_list_path=None) -> pd.DataFrame:
    """Compute percentage of overlap for each CNV with specific chromosomic regions: centromere and Segmental Duplication. By default the overlap is map on Hg19 Human genome.

    :param cnvs: list of CNVs with their scores
    :type cnvs: pd.DataFrame
    :param centromere_list_path: Pathway to a tsv file containing a list of centromere regions coordinates. By default will use the list human centromere mapped on Hg19 genome, defaults to None
    :type centromere_list_path: str, optional
    :param segdup_list_path: Pathway to a tsv file containing a list of Segmental duplication regions coordinates. By default will use the list human segmental duplication mapped on Hg19 genome, defaults to None
    :type segdup_list_path: str, optional
    :return: list of CNVs with chromosomic features columns aggregated
    :rtype: pd.DataFrame
    """
    this_dir, this_filename = split(__file__)
    if centromere_list_path == None:
        centromere_list_path = join(
            this_dir, "data", "Region_centromere_hg19.dat")
    if segdup_list_path == None:
        segdup_list_path = join(this_dir, "data", "SegDup_filtres_Ok_Oct.map")

    cnvs = getCentromereOverlap(cnvs, centromere_list_path)
    dc_logger.info("Centromere overlap added to CNVs")
    cnvs = getSegDupOverlap(cnvs, segdup_list_path)
    dc_logger.info("Both chromosomic annotation finished")
    return cnvs


def getCentromereOverlap(cnvs: pd.DataFrame, centromeres_list_path: str) -> pd.DataFrame:
    """Compute percentage of overlap for each CNV with centromere.

    :param cnvs: list of CNVs with their scores
    :type cnvs: pd.DataFrame
    :param centromeres_list_path: Pathway to a tsv file containing a list of centromere regions coordinates
    :type centromeres_list_path: str
    :raises Exception: If the given centromere list file pathway doesn't exist
    :raises Exception: If the given file hasn't mandatory columns (`CHR`, `START`, `STOP`)
    :return: list of CNVs with centromere overlap feature aggregated
    :rtype: pd.DataFrame
    """
    if exists(centromeres_list_path):
        centromeres = pd.read_csv(centromeres_list_path, sep='\t')
    else:
        raise Exception("File {} note found".format(centromeres_list_path))

    # validate the necessary titles of the regions of interest file
    nescessaryRegionTitles = ["CHR", "START", "STOP"]
    if len(centromeres.columns.intersection(nescessaryRegionTitles)) < 3:
        raise Exception(
            "The input file for the regions of interest must contain the following columns: CHR, START and STOP")

    centromeres.rename(columns={
                       "CHR": "CHR_centro", "START": "START_centro", "STOP": "STOP_centro"}, inplace=True)
    temp = pd.merge(cnvs, centromeres, left_on="CHR", right_on="CHR_centro")
    overlap = (temp[["STOP_centro", "STOP"]].min(
        axis=1) - temp[["START_centro", "START"]].max(axis=1) + 1)/(temp.STOP - temp.START + 1)
    overlap = np.where(overlap > 0, overlap, 0)
    dc_logger.info("Centromere overlap created")
    cnvs["overlapCNV_Centromere"] = overlap
    return cnvs


def getSegDupOverlap(cnvs: pd.DataFrame, segdup_list_path: str) -> pd.DataFrame:
    """Compute percentage of overlap for each CNV with Segmental Duplication regions.

    :param cnvs: list of CNVs with their scores
    :type cnvs: pd.DataFrame
    :param segdup_list_path: Pathway to a tsv file containing a list of centromere regions coordinates
    :type segdup_list_path: str
    :raises Exception:  If the given sgmental duplication list file pathway doesn't exist
    :return: list of CNVs with sgmental duplication overlap feature aggregated
    :rtype: pd.DataFrame
    """
    if exists(segdup_list_path):
        segdups = pd.read_csv(segdup_list_path, sep='\t')
        dc_logger.info("Segmental duplication file opened")
    else:
        raise Exception("File note found")
    complete_cnvs = pd.DataFrame()

    for chr in cnvs.CHR.unique():
        cnvs_chr = cnvs[cnvs.CHR == chr].copy()
        segdups_chr = segdups[segdups.CHR == chr].copy()
        segdups_chr.reset_index(inplace=True)
        vals = segdups_chr.apply(lambda x: computeOneOverlap(
            cnvs_chr, x.START, x.STOP), axis=1)
        overlaps = pd.DataFrame(vals.tolist())
        cnvs_chr.loc[:, "overlapCNV_SegDup"] = overlaps.sum().tolist()
        complete_cnvs = pd.concat([complete_cnvs, cnvs_chr])
    dc_logger.info("Segmental duplication overlap computed and CNVs annotated")
    return complete_cnvs


def computeOneOverlap(cnvs:pd.DataFrame, START:int, STOP:int):
    """Compute the percentage of overlap between a list of CNVs and START -- STOP coordinates of a given Segmental Duplication region. 
    Warning: CNVs must be on the same chromsome than the tested segmental duplication

    :param cnvs: list of CNVs coming from a unique chromosome similar to the segmental duplication analyzed 
    :type cnvs: pd.DataFrame
    :param START: Start coordinate of the segmental duplication region
    :type START: int
    :param STOP: Stop coordinate of the segmental duplication region
    :type STOP: int
    :return: An array listing the percentage of overlap for all given CNVs with the segmental duplication region
    :rtype: np.ArrayLike
    """
    overlap = (np.where(cnvs['STOP'] < STOP, cnvs['STOP'], STOP) - np.where(
        cnvs['START'] < START, START, cnvs['START']) + 1)/(cnvs['STOP'] - cnvs['START'] + 1)
    overlap = np.where(overlap > 0, overlap, 0)
    return overlap


def transformTwoAlgsFeatures(cnvs: pd.DataFrame) -> pd.DataFrame:
    """CHeck if the TwoAlgs column has the right format and correct it if not.

    :param cnvs: list of CNVs with their scores
    :type cnvs: pd.DataFrame
    :return: list of CNVs with the TwoAlgs column corrected if necessary
    :rtype: pd.DataFrame
    """
    if cnvs.TwoAlgs.dtype == object:
        cnvs.TwoAlgs = cnvs.TwoAlgs.str[:-1]
        cnvs.TwoAlgs = cnvs.TwoAlgs.astype(int)
    if cnvs.TwoAlgs.describe()[7] <= 1.0:
        cnvs.TwoAlgs = cnvs.TwoAlgs * 100
        dc_logger.info("Transform TwoAlgs function into percentage format")
    elif cnvs.TwoAlgs.describe()[7] > 1.0 <= 100.0:
        dc_logger.info("Keep TwoAlgs function into percentage format")
    else:
        dc_logger.info(
            "Error in TwoAlgs format must be a percentage or a rate")
        quit()
    return cnvs


def main():
    """ """
    cnvs = pd.read_csv('../data/UKBB_clean_for_DigCNV.tsv',
                       sep='\t', index_col=False)
    cnvs = getSegDupOverlap(cnvs, '../data/SegDup_filtres_Ok_Oct.map')
    print('ok')


if __name__ == '__main__':
    main()



from digcnv.digCNV_logger import logger as dc_logger

import tempfile
from os.path import split, join, exists
from subprocess import Popen, PIPE
import pandas as pd


def formatPennCNVforCNVision(pennCNVfile_path: str, output_path: str):
    """Format PennCNV file to be readable by CNVision algorithm and write formatted CNVs into the given output path

    :param pennCNVfile_path: Pathway of the PennCNV output file listing all CNVs with their score. For File format example read README.md page on DigCNV github page.
    :type pennCNVfile_path: str
    :param output_path: Pathway of the PennCNV file formatted to use in CNVision function.
    :type output_path: str

    """
    dc_logger.info("format and filter PennCNV file to CNVision requirements")
    this_dir = split(__file__)[0]
    perl_script = join(this_dir, 'data', "CNVision_format_merge_ukbb.pl")
    proc1 = Popen(["perl", perl_script, "--PNformat", pennCNVfile_path,
                  output_path, 'tmp'], stdout=PIPE,  bufsize=1, text=True)
    while proc1.poll() is None:
        line = proc1.stdout.readline()
        dc_logger.info(line.rstrip())
    proc1.wait()
    dc_logger.info("PennCNV file formatted for CNVision")


def formatQuantiSNPforCNVision(quantiSNP_file_path: str, output_path: str):
    """Format QuantiSNP file to be readable by CNVision algorithm and write formatted CNVs into the given output path

    :param quantiSNP_file_path: Pathway of the QuantiSNP output file listing all CNVs with their score. For File format example read README.md page on DigCNV github page.
    :type quantiSNP_file_path: str
    :param output_path: Pathway of the QuantiSNP file formatted to use in CNVision function.
    :type output_path: str
    """
    dc_logger.info("format and filter QuantiSNP file to CNVision requirements")
    this_dir = split(__file__)[0]
    perl_script = join(this_dir, 'data', "CNVision_format_merge_ukbb.pl")
    proc1 = Popen(["perl", perl_script, "--QTformat", quantiSNP_file_path,
                  output_path, 'tmp'], stdout=PIPE, bufsize=1, text=True)
    while proc1.poll() is None:
        line = proc1.stdout.readline()
        dc_logger.info(line.rstrip())
    proc1.wait()
    dc_logger.info("PennCNV file formatted for CNVision")


def runMergingScript(formated_data_paths: list, output_path: str):
    """Create a merged CNV list file at the given output path. CNV files inputed must be in CNVision ready format.

    :param formated_data_paths: list of pathways to called CNVs and formatted for CNVision.
    :type formated_data_paths: list
    :param output_path: Pathway where the merged list of CNV will be written with the new algorithm overlap information added.
    :type output_path: str
    """
    dc_logger.info("format and filter QuantiSNP file to CNVision requirements")
    this_dir = split(__file__)[0]
    perl_script = join(this_dir, 'data', "CNVision_format_merge_ukbb.pl")
    proc1 = Popen(["perl", perl_script, "--merge", formated_data_paths[0],
                  formated_data_paths[1], output_path, 'tmp'], stdout=PIPE, bufsize=1, text=True)
    while proc1.poll() is None:
        line = proc1.stdout.readline()
        dc_logger.info(line.rstrip())
    proc1.wait()
    dc_logger.info("PennCNV file formatted for CNVision")


def mergeMultipleCNVCallingOutputs(list_calling_outputs_path: list, list_calling_softwares: list) -> pd.DataFrame:
    """Create a dataframe containing merged CNVs comming from the given algorithm outputs. Presently working only with PennCNV and QuantiSNP only.
    Please indicate the list of pathway and the list of algorithm names in the same order.

    :param list_calling_outputs_path: list of calling output pathways
    :type list_calling_outputs_path: list_calling_outputs_path: list
    :param list_calling_softwares: {`PennCNV`, `QuantiSNP`} list of calling algorithm names, indicated in the same order as the list of pathways 
    :type list_calling_softwares: list_calling_softwares: list
    :returns: Dataframe containing all CNVs coming from all caller outputs merged and with their overlap annotated
    :rtype: pd.DataFrame
    :raises Exception: If two given list haven't same sizes
    :raises Exception: If the calling algorithm name isn't supported

    """
    if len(list_calling_outputs_path) != len(list_calling_softwares):
        raise Exception("Both list must have same sizes")
    list_tmp_paths = []
    with tempfile.TemporaryDirectory() as tmp_dir:
        for i, calling_soft in enumerate(list_calling_softwares):
            if calling_soft == 'PennCNV':
                formatPennCNVforCNVision(list_calling_outputs_path[i], tmp_dir)
                if exists(join(tmp_dir, 'PC_tmp_CNVisionFormated.txt')):
                    list_tmp_paths.append(
                        join(tmp_dir, 'PC_tmp_CNVisionFormated.txt'))
            elif calling_soft == 'QuantiSNP':
                formatQuantiSNPforCNVision(
                    list_calling_outputs_path[i], tmp_dir)
                if exists(join(tmp_dir, 'QS_tmp_CNVisionFormated.txt')):
                    list_tmp_paths.append(
                        join(tmp_dir, 'QS_tmp_CNVisionFormated.txt'))
            else:
                raise Exception(
                    "Given calling software: {} in't support by the software please contact us if you want to add this software".format(calling_soft))
        runMergingScript(list_tmp_paths, tmp_dir)
        CNVs_list = pd.read_csv(
            join(tmp_dir, 'Sum_CNVisionMerged_PC_QS_tmp.txt'), sep='\t')

    return CNVs_list


def main():
    print(mergeMultipleCNVCallingOutputs(['/home/thomas/Documents/temp_data/PC_allCNV.txt',
          '/home/thomas/Documents/temp_data/QS_allCNV.txt'], ['PennCNV', 'QuantiSNP']).head())


if __name__ == "__main__":
    main()
